from adls_management.utils import messages
from adls_management.interface_file_management import interface_file_handling
from datetime import datetime, date
from uuid import uuid4
from adls_management.adls_management import folder_management
from os import path


class FilesBatchPhaseFromTo:
    """
        Convenience modules to move files from one stage to the next.
        locations need to be configured in resources/locations.json (default)
    """
    right_now = datetime.now().isoformat(timespec="microseconds").replace(":", "-")
    # todays_date = datetime.today().strftime("%Y-%m-%d")

    def __init__(self, configuration_file, run_id=None):
        self.run_id = None
        self.run_id, self.time_id = self.determine_run_id()
        print("run_id is >%s< with time_stamp >%s<" % (self.run_id, self.time_id))
        self.result = messages.message["ok"]
        self.file_handler = interface_file_handling.InterfaceFileHandling(configuration_file=configuration_file)
        self.mgmt = folder_management.ADLSFolderManagement(configuration_file=configuration_file)
        self.container_handler = self.mgmt.container_client
        self.oldest_name = None

    def determine_run_id(self):
        if self.run_id is None:
            the_run_id = str(uuid4())
        else:
            the_run_id = self.run_id
        return the_run_id, self.right_now

    def determine_target_name(self, base_name):
        return path.join(base_name, self.right_now + "--" + self.run_id)

    def determine_oldest_folder(self, base_folder):
        # given the fact that this code also created the folder structure:
        # all folders have timestamp(formatted)+ '--' + uuid4
        # NOTE: A directory is listed by name on Azur Blobs, so actually the first one found is the oldest,
        #       given the naming convention of this tooling
        # print("content of %s:" % base_folder)
        oldest_name = "not_found"
        # introduce the year 3030 problem:
        oldest = 30301224093810129694
        result, sources, source_names = self.file_handler.list_files(location=base_folder, file_pattern=".*json")
        # for folder in self.container_handler.walk_blobs(base_folder, delimiter='/'):
        for folder in source_names:
            # print(folder)
            if "--" not in folder:
                # print(">%s< is not a file we are interested in" % folder)
                continue
            time_from_name = folder.split("--")[0]
            file_nr_as_str = ''.join(c for c in time_from_name if c.isdigit())
            # print("file >%s< has timestamp >%s< which is >%s<" % (folder, time_from_name, file_nr_as_str))
            if file_nr_as_str == "":
                self.oldest_name = "not_found"
            else:
                file_nr = int(file_nr_as_str)
                if file_nr < oldest:
                    self.oldest_name = folder
                    oldest = file_nr

        return oldest, self.oldest_name

    def from_incoming2todo(self):
        """
            Move the files to 'todo' so they can be processed. The 'incoming' is freed-up for new files.
        """
        print("=== from_incoming2todo ===")
        result = self.from_to(self.file_handler.settings.incoming, self.file_handler.settings.todo)
        result["reference"] = "from >%s< to >%s<" % (self.file_handler.settings.incoming, self.file_handler.settings.todo)
        return result

    def from_todo2busy(self):
        """
            Move the files to 'busy' as pre-processing step of the actual processing
            Move the files from the oldest folder to the 'busy' folder
            Note: In this release, we cannot have multiple processes processing files.
                The 'busy' folder can only contain one single set of files
        """
        # find the oldest folder
        print("=== from_todo2busy ===")
        from_folder = self.file_handler.settings.todo
        print("from_folder is >%s<" % from_folder)
        oldest, self.oldest_name = self.determine_oldest_folder(base_folder=from_folder)
        print("oldest_folder is >%s<" % self.oldest_name)
        if self.oldest_name == "not_found" or self.oldest_name == "" or self.oldest_name is None:
            result = messages.message["could_not_determine_oldest_folder"]
            return result

        # to_work_on_folder = oldest_name.split('/')[1]
        target = self.file_handler.settings.busy
        result = self.from_to(path.join(from_folder, self.oldest_name), target)
        if result["code"] == "OK":
            result["reference"] = "oldest_name >%s< went to >%s<" % (self.oldest_name, target)
        if result["code"] == "MLU-FH-009":
            # empty source folder
            result = self.mgmt.create_directory(path.join(self.file_handler.settings.done, self.oldest_name))
            if result["code"] == "OK":
                result = self.mgmt.delete_directory(directory=path.join(self.file_handler.settings.todo, self.oldest_name))
        return result

    def from_busy2done(self):
        print("=== from busy2done ===")
        from_folder = self.file_handler.settings.busy
        print("from_folder is >%s<" % from_folder)
        oldest, self.oldest_name = self.determine_oldest_folder(base_folder=from_folder)
        print("oldest_folder is >%s<" % self.oldest_name)
        if self.oldest_name == "not_found" or self.oldest_name == "" or self.oldest_name is None:
            result = messages.message["could_not_determine_oldest_folder"]
            return result

        from_folder = path.join(self.file_handler.settings.busy, self.oldest_name)
        to_folder = path.join(self.file_handler.settings.done, self.oldest_name)
        print("from >%s< to >%s<" % (from_folder, to_folder))
        result = self.from_to(from_location=from_folder, to_location=to_folder)
        return result

    def from_done2hist(self):
        """
            Moves files from 'done' to 'hist', where 'hist' could be a cold location
            Use this on a regular basis to clean-up the 'done' folder
        """
        print("=== from done2hist ===")
        result = self.from_to(self.file_handler.settings.done, self.file_handler.settings.hist)
        return result

    def from_busy2redo(self):
        """
            When an error occurred use this method to move the files to the configured 'redo'
        """
        print("=== from busy2redo ===")
        result = self.from_to(self.file_handler.settings.busy, self.file_handler.settings.redo)
        return result

    def from_to(self, from_location, to_location):
        """
            Move files from the configured from_location to the configured to_location (check locations.json)
        """
        print("--- from_to ---")
        result, files, file_names = self.file_handler.list_files(location=from_location, file_pattern=".*json")
        if result["code"] == "OK":
            if len(files) > 0:
                target_directory = self.determine_target_name(to_location)
                result = self.mgmt.create_directory(directory=target_directory)
                if result["code"] == "OK":
                    result = self.free_from_to(from_location=from_location, to_location=target_directory)
            else:
                print("No files in source:", from_location)
                result = messages.message["no_files_in_source"]
        return result

    def free_from_to(self, from_location, to_location):
        """
        """
        print("free - from >%s< to >%s<" % (from_location, to_location))
        result = self.file_handler.move_files(from_location=from_location
                                                  , to_location=to_location
                                                  , file_pattern=".*json")
        return result



from datetime import datetime

from adls_management.adls_management import connection
from adls_management.utils import settings, messages
import os
from azure.core import exceptions
from azure.storage.blob import BlobClient, BlobLeaseClient
import time
from adls_management.adls_management import folder_management
import re


class InterfaceFileHandling:
    right_now = datetime.now().isoformat(timespec="microseconds").replace(":", "-")

    def __init__(self, configuration_file):
        # use provided connection config file
        self.settings = settings.GenericSettings(configuration_file=configuration_file)
        self.settings.get_config()

        self.service_client, self.file_system_client, self.blob_service_client, self.container_client, self.sas_token \
            = connection.ConnectionManagement(self.settings).create_connection(
            storage_account_name=self.settings.storage_account_name
            , storage_account_key=self.settings.storage_account_key
            , container=self.settings.storage_container
        )
        self.mgmt = folder_management.ADLSFolderManagement(configuration_file=configuration_file)
        self.max_wait_in_sec = 60
        self.recheck = 10
        self.tgt = None

    def move_files(self, from_location, to_location, file_pattern):
        result = self.copy_files(from_location=from_location, to_location=to_location, file_pattern=file_pattern)
        if result == messages.message["ok"]:
            result = self.remove_files(location=from_location, file_pattern=file_pattern)
            if result["code"] == "OK":
                if from_location != self.mgmt.settings.incoming:
                    # don't remove the base incoming directory. This is the only directory that does not have subdirs
                    result = self.mgmt.delete_directory(from_location)

        return result

    def copy_files(self, from_location, to_location, file_pattern):
        result = messages.message["copy_files_failed"]
        result, sources, source_names = self.list_files(location=from_location, file_pattern=file_pattern)
        if sources is None:
            print("no files found in source >" + from_location + "<.")
            result = messages.message["ok"]
            result["reference"] = "No files in source: " + from_location
            return result

        for file in sources:
            # https://docs.microsoft.com/en-us/azure/storage/blobs/storage-blob-copy?tabs=python
            src_file = file # os.path.basename(file)
            tgt_file = to_location + "/" + os.path.basename(src_file)
            src_blob = BlobClient(
                self.blob_service_client.url,
                container_name=self.settings.storage_container,
                blob_name=src_file,
                credential=self.sas_token
            )
            lease = BlobLeaseClient(src_blob)
            lease.acquire()
            source_props = src_blob.get_blob_properties()
            print("Lease state for source file %s: %s" %(src_file ,source_props.lease.state))

            print(f"Copying %s.%s to %s using url %s"
                  % (self.settings.storage_container, src_file, tgt_file, src_blob.url))

            self.tgt = self.blob_service_client.get_blob_client(self.settings.storage_container, tgt_file)

            # download_file_path = os.path.join(".", str.replace("tryout", '.txt', 'DOWNLOAD.txt'))
            # print("\nDownloading blob to \n\t" + download_file_path)
            # with open(download_file_path, "wb") as download_file:
            #    download_file.write(src_blob.download_blob().readall())

            try:
                action = self.tgt.start_copy_from_url(src_blob.url)
                copy_id = action["copy_id"]
                status = action["copy_status"]
                error = action["error_code"]
                self.wait_condition(condition=self.check_copy_status
                                    , timeout=self.max_wait_in_sec
                                    , granularity=self.recheck)

                properties = self.tgt.get_blob_properties()
                print("Total bytes: " + str(properties.size))
                copy_props = properties.copy
                if copy_props["status"] != "success":
                    self.tgt.abort_copy(copy_id=copy_props["id"])
                    print(f"Unable to copy blob %s to %s. Status: %s" % (src_file, tgt_file, copy_props.status))
                    result = messages.message["copy_files_failed"]
                    break
                    # Note: We do not release the lease in case of errors

                if source_props.lease.state == "leased":
                    # Break the lease on the source blob.
                    lease.break_lease()
                    # Update the source blob's properties to check the lease state.
                    source_props = src_blob.get_blob_properties()
                    print("Lease state: " + source_props.lease.state)

                result = messages.message["ok"]
            except exceptions.ResourceNotFoundError as e:
                print("Azure reported a resource not found error: ", e)
                result = messages.message["resource_not_found"]
                result["reference"] = f"source: %s, targte: %s" % (src_file, tgt_file)

        return result

    def wait_condition(self, condition, timeout=10.0, granularity=1.0, time_factory=time):
        """
        thankfully re-used from
        https://stackoverflow.com/questions/45455898/polling-for-a-maximum-wait-time-unless-condition-in-python-2-7
        """
        end_time = time.time() + timeout  # compute the maximal end time
        status = condition()  # first condition check, no need to wait if condition already True
        while not status and time.time() < end_time:  # loop until the condition is false and timeout not exhausted
            time.sleep(granularity)  # release CPU cycles
            status = condition()  # check condition
        return status

    def check_copy_status(self):
        properties = self.tgt.get_blob_properties()
        copy_props = properties.copy
        print("properties: ", copy_props)
        # Display the copy status
        print("Copy status: " + copy_props["status"])
        # print("Copy progress: " + copy_props["progress"])
        # print("Completion time: " + str(copy_props["completion_time"]))
        if copy_props["status"] == "pending":
            done = False
        else:
            done = True
        return done

    def check_files(self, source_location, target_location, file_pattern):
        """
            source and target must have the same files (simple list comparison)
        """
        result, source_list, source_names = self.list_files(location=source_location, file_pattern=file_pattern)
        result, target_list, target_names = self.list_files(location=target_location, file_pattern=file_pattern)
        if source_names == target_names:
            result = messages.message["ok"]
            result["reference"] = "location >" + source_location \
                                  + "< and location >" + target_location \
                                  + "< have the same files."
        else:
            result = messages.message["check_files_difference_found"]
            result["reference"] = "location >" + source_location \
                                  + "< and location >" + target_location \
                                  + "< do NOT have the same files."
        return result

    def list_files(self, location, file_pattern):
        files = []
        file_names = []
        if file_pattern is not None:
            pattern = re.compile(file_pattern)
        else:
            pattern = None
        try:
            paths = self.file_system_client.get_paths(path=location)
            for path in paths:
                if pattern is None:
                    files.append(path.name)
                    file_names.append(path.name.split('/')[1])
                else:
                    if pattern.match(path.name) is not None:
                        files.append(path.name)
                        file_names.append(path.name.split('/')[1])
            result = messages.message["ok"]
        except Exception as e:
            print(e)
            result = messages.message["list_directory_error"]
            result["reference"] = "directory: " + location + " - " + str(e)
            return result, None, None
        return result, files, file_names

    def historize_files(self, source_location, file_pattern, recursive=True):
        return

    def remove_files(self, location, file_pattern):
        result, sources, source_names = self.list_files(location=location, file_pattern=file_pattern)
        if result["code"] == "OK":
            for file in sources:
                try:
                    src_blob = BlobClient(
                        self.blob_service_client.url,
                        container_name=self.settings.storage_container,
                        blob_name=file,
                        credential=self.sas_token
                    )
                    src_blob.delete_blob(delete_snapshots="include")
                    print("blob removed:", file)
                except Exception as e:
                    print("Exception on blob removal:", e)
                    result = messages.message["remove_files_failed"]
        return result

    def upload_file(self, location, filename):
        if '/' in filename:
            tgt_filename = filename.split('/')[1]
        else:
            tgt_filename = filename

        tgt_blob = BlobClient(
            self.blob_service_client.url,
            container_name=self.settings.storage_container,
            blob_name=os.path.join(location, tgt_filename),
            credential=self.sas_token
        )

        with open(filename, "rb") as data:
            response = tgt_blob.upload_blob(data, overwrite=True)
            print("upload result: ", response["error_code"])
            if response["error_code"] is None:
                return messages.message["ok"]
            else:
                return messages.message["upload_failed"]

    def download_file(self, location, filename):

        src_blob = BlobClient(
            self.blob_service_client.url,
            container_name=self.settings.storage_container,
            blob_name=filename,
            credential=self.sas_token
        )

        path_to_file = os.path.dirname(filename)
        download_path_for_file = os.path.join(self.settings.download_location, path_to_file)
        os.makedirs(download_path_for_file, exist_ok=True)
        download_file_path = os.path.join(self.settings.download_location, filename)
        print("\nDownloading blob to \n\t" + download_file_path)

        with open(download_file_path, "wb") as download_file:
            download_file.write(src_blob.download_blob().readall())
        return

    def download_files(self, location=None, pattern=".*.json"):

        if location is None:
            location = self.settings.busy
        result, files, file_names = self.list_files(location=location, file_pattern=pattern)
        if result["code"] == "OK":
            for file in files:
                print("Downloading file >%s< from location >%s<" % (file, location))
                self.download_file(location=location, filename=file)
        else:
            print("An error occurred listing files", result)



message = {
    "MLU_GEN_GROUP": {
        "code": "MLU_GEN"
        , "message": "Group for generic error or warning regarding Metadata Lake Utilities"
        , "level": "INFO"
        , "reference": "Info about location and/or values that lead to the error (non-PII)"
    },
    "ok": {
        "code": "OK"
        , "message": "No errors encountered"
        , "level": "INFO"
        , "reference": "None"
    },
    "not_implemented": {
        "code": "MLU-GEN-000"
        , "message": "Function not yet implemented. Please contact the development team and mention code 'MLU-GEN-000'."
        , "level": "ERROR"
        , "reference": "None"
    },
    "not_found": {
        "code": "MLU-GEN-001"
        , "message": "Could not find what we were looking for"
        , "level": "ERROR"
        , "reference": "None"
    },
    "os_error": {
        "code": "MLU-GEN-002"
        , "message": "An OS error occurred"
        , "level": "ERROR"
        , "reference": "None"
    },
    "undetermined": {
        "code": "MLU-GEN-003"
        , "message": "An unexpected error occurred. Please contact the development team."
        , "level": "ERROR"
        , "reference": "None"
    },
    "main_config_not_found": {
        "code": "MLU-GEN-004"
        , "message": "Main configuration file not found"
        , "level": "FATAL"
        , "reference": "None"
    },
    "azure_secrets_not_found": {
        "code": "MLU-GEN-005"
        , "message": "azure secrets file not found"
        , "level": "FATAL"
        , "reference": "None"
    },
    "file_locations_not_found": {
        "code": "MLU-GEN-006"
        , "message": "file_locations file not found"
        , "level": "ERROR"
        , "reference": "None"
    },
    #
    # file file_handler messages
    #
    "list_directory_error": {
        "code": "MLU-FH-001"
        , "message": "Could not list content of the directory"
        , "level": "ERROR"
        , "reference": "None"
    },
    "copy_files_failed": {
        "code": "MLU-FH-002"
        , "message": "Could not copy files"
        , "level": "ERROR"
        , "reference": "None"
    },
    "remove_files_failed": {
        "code": "MLU-FH-003"
        , "message": "Could not remove files"
        , "level": "ERROR"
        , "reference": "None"
    },
    "check_files_difference_found": {
        "code": "MLU-FH-004"
        , "message": "Differences found"
        , "level": "ERROR"
        , "reference": "None"
    },
    "resource_not_found": {
        "code": "MLU-FH-005"
        , "message": "Resource could not be found"
        , "level": "ERROR"
        , "reference": "None"
    },
    "upload_failed": {
        "code": "MLU-FH-006"
        , "message": "File could not be uploaded"
        , "level": "ERROR"
        , "reference": "None"
    },
    "download_failed": {
        "code": "MLU-FH-007"
        , "message": "File could not be downloaded"
        , "level": "ERROR"
        , "reference": "None"
    },
    "could_not_determine_oldest_folder": {
        "code": "MLU-FH-008"
        , "message": "Could not determine the oldest (based on folder name) folder"
        , "level": "ERROR"
        , "reference": "None"
    },
    "no_files_in_source": {
        "code": "MLU-FH-009"
        , "message": "No files found in folder"
        , "level": "WARNING"
        , "reference": "None"
    },
    #
    # Security related messages
    #
    "unsupported_meta_version_azure_secrets": {
        "code": "MLU-SEC-001"
        , "message": "Incorrect meta_version used in secrets file"
        , "level": "FATAL"
        , "reference": "None"
    },
    #
    # directory handling messages
    #
    "directory_creation_error": {
        "code": "MLU-DIR-001"
        , "message": "Directory could not be created"
        , "level": "ERROR"
        , "reference": "None"
    },
    "directory_removal_error": {
        "code": "MLU-DIR-002"
        , "message": "Directory could not be removed"
        , "level": "ERROR"
        , "reference": "None"
    }
}
